/**
  Copyright 2016 Olivier blunt (business@blunt.sh)

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

	  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
'use strict';

const fs = require('fs-ext');

/**
 * Read our lock file's data
 * @param	file	file
 * @return	object
 */
function readData(file)
{
	// Read
	var buffer = new Buffer(2048);
	var lg = fs.readSync(file, buffer, 0, 2048);
	var content = buffer.toString('utf8', 0, lg);
	
	// Parse
	try {
		content = JSON.parse(content);
	} catch (err) {
		return {};
	}
	return content || {};
}

/**
 * Save some data to the lock file
 * @param	file	file
 * @param	object	data
 */
function writeData(file, data)
{
	try {
		// Encode
		data = JSON.stringify(data);
		
		// Write
		fs.ftruncateSync(file, 0);
		fs.writeSync(file, data, 0);
	} catch (err) {
		console.error(err);
	}
}

/**
 * See readme.md for usage
 */
module.exports = function (gulp, options)
{
	// Default options
	options = options || {};
	options.afk_ttl = options.afk_ttl || 600;
	options.lock_path = options.lock_path || process.cwd()+'/gulp.lock';
	
	// Out data stored in the lock file
	var lock_data;
	
	// Our locked file descriptor
	var lock;
	try {
		lock = fs.openSync(options.lock_path, 'r+');
		
		// Read lockfile content
		lock_data = readData(lock);
		
		// We are not in our gulp service
		if (lock_data.pid != process.pid) {
			// Create a task to kill our gulp service
			gulp.task('waitforit:kill', function (cb)
			{
				console.log('Killing...');
				process.kill(lock_data.pid, 'SIGINT');
				cb();
			});
			//console.log('-- another gulp process: ', lock_data);
			return;
		}
	} catch (er) {
		//console.log('-- no pid file: ' + er);
		return;
	}
	
	// True if the file is already locked
	var locked = false;
	
	// List current running tasks
	var running_tasks = {};
	
	// Unlock timer to prevent useless unlock/lock
	var unlock_timer_id;
	
	// Afk kill timer
	var afk_timer_id;
	
	/**
	 * Add the new task to running_tasks
	 * @param	Event	ev
	 */
	function onStart(ev)
	{
		running_tasks[ev.task] = true;
		
		// Lock
		if (!locked) {
			//console.log('-- lock');
			locked = true;
			try {
				fs.flockSync(lock, 'ex');
			} catch (err) {
				locked = false;
				console.error('gulp-waitforit: Failed to lock: '+err);
			}
		}
		
		// Remove afk timer
		if (afk_timer_id) {
			clearTimeout(afk_timer_id);
			afk_timer_id = null;
		}
	}
	
	/**
	 * Occurs when a task stop or fails
	 * @param	Event	ev
	 */
	function onStop(ev)
	{
		delete running_tasks[ev.task];
		
		if (ev.err) {
			// Add error to lock file before we unlock
			if (!lock_data.errors) lock_data.errors = {};
			lock_data.errors[ev.task] = ev.err.stack;
			writeData(lock, lock_data);
		} else if (lock_data.errors && lock_data.errors[ev.task]) {
			// Remove error from lock file before we unlock
			delete lock_data.errors[ev.task];
			writeData(lock, lock_data);
		}
		
		// Unlock in 50ms
		if (unlock_timer_id) clearTimeout(unlock_timer_id);
		unlock_timer_id = setTimeout(unlock, 50);
		
		// Put back our afk timer
		afk_timer_id = setTimeout(afkKill, options.afk_ttl * 1000);
	}
	
	/**
	 * Unlock if there is no more running tasks
	 */
	function unlock()
	{
		// Still no more running tasks
		if (Object.keys(running_tasks).length == 0 && locked) {
			//console.log('-- unlock');
			try {
				fs.flockSync(lock, 'un');
				locked = false;
			} catch (err) {
				console.error('gulp-waitforit: Failed to unlock: '+err);
			}
		}
	}
	
	/** Just kill our process */
	function afkKill()
	{
		console.log('Im bored, bye.');
		process.exit(0);
	}
	
	// Register events
	gulp.on('task_start', onStart);
	gulp.on('task_stop', onStop);	
	gulp.on('task_err', onStop);
	
	// Global errors
	process.on('uncaughtException', function (exception)
	{
		// Add error to lock file before we unlock
		if (!lock_data.errors) lock_data.errors = {};
		lock_data.errors.global_error = exception.stack;
		writeData(lock, lock_data);
		
		throw exception;
	});
	// Remove last global error
	if (lock_data.errors && lock_data.errors.global_error) {
		delete lock_data.errors.global_error;
		writeData(lock, lock_data);
	}
	
	//TODO: detect gulpfile.js changes to quit
};